import binascii
from decimal import Decimal
from typing import Optional, Union

from .. import xdr as stellar_xdr
from ..muxed_account import MuxedAccount
from ..utils import raise_if_not_valid_amount, raise_if_not_valid_hash
from .operation import Operation

__all__ = ["LiquidityPoolWithdraw"]


class LiquidityPoolWithdraw(Operation):
    """The :class:`LiquidityPoolWithdraw` object, which represents a LiquidityPoolWithdraw
    operation on Stellar's network.

    Creates a liquidity pool withdraw operation.

    Threshold: Medium

    See `Liquidity Pool Withdraw <https://developers.stellar.org/docs/start/list-of-operations/#liquidity-pool-withdraw>`_ for more information.

    :param liquidity_pool_id: The liquidity pool ID.
    :param amount: Amount of pool shares to withdraw.
    :param min_amount_a: Minimum amount of first asset to withdraw.
    :param min_amount_b: Minimum amount of second asset to withdraw.
    :param source: The source account for the operation. Defaults to the
        transaction's source account.
    """

    _XDR_OPERATION_TYPE: stellar_xdr.OperationType = (
        stellar_xdr.OperationType.LIQUIDITY_POOL_WITHDRAW
    )

    def __init__(
        self,
        liquidity_pool_id: str,
        amount: Union[str, Decimal],
        min_amount_a: Union[str, Decimal],
        min_amount_b: Union[str, Decimal],
        source: Optional[Union[MuxedAccount, str]] = None,
    ):
        super().__init__(source)
        self.liquidity_pool_id: str = liquidity_pool_id
        self.amount: str = str(amount)
        self.min_amount_a: str = str(min_amount_a)
        self.min_amount_b: str = str(min_amount_b)
        raise_if_not_valid_amount(self.amount, "amount")
        raise_if_not_valid_amount(self.min_amount_a, "min_amount_a")
        raise_if_not_valid_amount(self.min_amount_b, "min_amount_b")
        raise_if_not_valid_hash(self.liquidity_pool_id, "liquidity_pool_id")

    def _to_operation_body(self) -> stellar_xdr.OperationBody:
        liquidity_pool_id_bytes = binascii.unhexlify(self.liquidity_pool_id)
        liquidity_pool_id = stellar_xdr.PoolID.from_xdr_bytes(liquidity_pool_id_bytes)
        amount = stellar_xdr.Int64(Operation.to_xdr_amount(self.amount))
        min_amount_a = stellar_xdr.Int64(Operation.to_xdr_amount(self.min_amount_a))
        min_amount_b = stellar_xdr.Int64(Operation.to_xdr_amount(self.min_amount_b))
        liquidity_pool_withdraw_op = stellar_xdr.LiquidityPoolWithdrawOp(
            liquidity_pool_id, amount, min_amount_a, min_amount_b
        )
        body = stellar_xdr.OperationBody(
            type=self._XDR_OPERATION_TYPE,
            liquidity_pool_withdraw_op=liquidity_pool_withdraw_op,
        )
        return body

    @classmethod
    def from_xdr_object(
        cls, xdr_object: stellar_xdr.Operation
    ) -> "LiquidityPoolWithdraw":
        """Creates a :class:`LiquidityPoolWithdraw` object from an XDR Operation object."""
        source = Operation.get_source_from_xdr_obj(xdr_object)
        assert xdr_object.body.liquidity_pool_withdraw_op is not None
        liquidity_pool_id_bytes = (
            xdr_object.body.liquidity_pool_withdraw_op.liquidity_pool_id.pool_id.hash
        )
        liquidity_pool_id = liquidity_pool_id_bytes.hex()
        amount = Operation.from_xdr_amount(
            xdr_object.body.liquidity_pool_withdraw_op.amount.int64
        )
        min_amount_a = Operation.from_xdr_amount(
            xdr_object.body.liquidity_pool_withdraw_op.min_amount_a.int64
        )
        min_amount_b = Operation.from_xdr_amount(
            xdr_object.body.liquidity_pool_withdraw_op.min_amount_b.int64
        )

        op = cls(
            source=source,
            liquidity_pool_id=liquidity_pool_id,
            amount=amount,
            min_amount_a=min_amount_a,
            min_amount_b=min_amount_b,
        )
        return op

    def __repr__(self):
        return (
            f"<LiquidityPoolWithdraw [liquidity_pool_id={self.liquidity_pool_id}, amount={self.amount}, "
            f"min_amount_a={self.min_amount_a}, min_amount_b={self.min_amount_b}, source={self.source}]>"
        )



from decimal import Decimal
from typing import List, Union

from ...asset import Asset
from ...call_builder.base.base_call_builder import BaseCallBuilder
from ...utils import convert_assets_to_horizon_param

__all__ = ["BaseStrictReceivePathsCallBuilder"]


class BaseStrictReceivePathsCallBuilder(BaseCallBuilder):
    """Creates a new :class:`StrictReceivePathsCallBuilder` pointed to server defined by horizon_url.

    The Stellar Network allows payments to be made across assets through path payments. A path payment specifies a
    series of assets to route a payment through, from source asset (the asset debited from the payer) to destination
    asset (the asset credited to the payee).

    A path search is specified using:

    - The source address or source assets.
    - The asset and amount that the destination account should receive.

    As part of the search, horizon will load a list of assets available to the
    source address and will find any payment paths from those source assets to
    the desired destination asset. The search's amount parameter will be used to
    determine if there a given path can satisfy a payment of the desired amount.

    If a list of assets is passed as the source, horizon will find any payment
    paths from those source assets to the desired destination asset.

    See `List Strict Receive Payment Paths <https://developers.stellar.org/api/aggregations/paths/strict-receive/>`__ for more information.

    :param source: The sender's account ID or a list of Assets. Any returned path must use a source that the sender can hold.
    :param destination_asset: The destination asset.
    :param destination_amount: The amount, denominated in the destination asset, that any returned path should be able to satisfy.
    :param horizon_url: Horizon server URL.
    """

    def __init__(
        self,
        source: Union[str, List[Asset]],
        destination_asset: Asset,
        destination_amount: Union[str, Decimal],
        **kwargs
    ) -> None:
        super().__init__(**kwargs)
        self.endpoint: str = "paths/strict-receive"

        if isinstance(destination_amount, Decimal):
            destination_amount = str(destination_amount)

        params = {
            "destination_amount": destination_amount,
            "destination_asset_type": destination_asset.type,
            "destination_asset_code": (
                None if destination_asset.is_native() else destination_asset.code
            ),
            "destination_asset_issuer": destination_asset.issuer,
        }
        if isinstance(source, str):
            params["source_account"] = source
        else:
            params["source_assets"] = convert_assets_to_horizon_param(source)

        self._add_query_params(params)



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .base import String

__all__ = ["SCSpecTypeUDT"]


class SCSpecTypeUDT:
    """
    XDR Source Code::

        struct SCSpecTypeUDT
        {
            string name<60>;
        };
    """

    def __init__(
        self,
        name: bytes,
    ) -> None:
        self.name = name

    def pack(self, packer: Packer) -> None:
        String(self.name, 60).pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> SCSpecTypeUDT:
        name = String.unpack(unpacker)
        return cls(
            name=name,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> SCSpecTypeUDT:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> SCSpecTypeUDT:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash((self.name,))

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.name == other.name

    def __repr__(self):
        out = [
            f"name={self.name}",
        ]
        return f"<SCSpecTypeUDT [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .account_entry import AccountEntry
from .claimable_balance_entry import ClaimableBalanceEntry
from .config_setting_entry import ConfigSettingEntry
from .contract_code_entry import ContractCodeEntry
from .contract_data_entry import ContractDataEntry
from .data_entry import DataEntry
from .ledger_entry_type import LedgerEntryType
from .liquidity_pool_entry import LiquidityPoolEntry
from .offer_entry import OfferEntry
from .trust_line_entry import TrustLineEntry
from .ttl_entry import TTLEntry

__all__ = ["LedgerEntryData"]


class LedgerEntryData:
    """
    XDR Source Code::

        union switch (LedgerEntryType type)
            {
            case ACCOUNT:
                AccountEntry account;
            case TRUSTLINE:
                TrustLineEntry trustLine;
            case OFFER:
                OfferEntry offer;
            case DATA:
                DataEntry data;
            case CLAIMABLE_BALANCE:
                ClaimableBalanceEntry claimableBalance;
            case LIQUIDITY_POOL:
                LiquidityPoolEntry liquidityPool;
            case CONTRACT_DATA:
                ContractDataEntry contractData;
            case CONTRACT_CODE:
                ContractCodeEntry contractCode;
            case CONFIG_SETTING:
                ConfigSettingEntry configSetting;
            case TTL:
                TTLEntry ttl;
            }
    """

    def __init__(
        self,
        type: LedgerEntryType,
        account: AccountEntry = None,
        trust_line: TrustLineEntry = None,
        offer: OfferEntry = None,
        data: DataEntry = None,
        claimable_balance: ClaimableBalanceEntry = None,
        liquidity_pool: LiquidityPoolEntry = None,
        contract_data: ContractDataEntry = None,
        contract_code: ContractCodeEntry = None,
        config_setting: ConfigSettingEntry = None,
        ttl: TTLEntry = None,
    ) -> None:
        self.type = type
        self.account = account
        self.trust_line = trust_line
        self.offer = offer
        self.data = data
        self.claimable_balance = claimable_balance
        self.liquidity_pool = liquidity_pool
        self.contract_data = contract_data
        self.contract_code = contract_code
        self.config_setting = config_setting
        self.ttl = ttl

    def pack(self, packer: Packer) -> None:
        self.type.pack(packer)
        if self.type == LedgerEntryType.ACCOUNT:
            if self.account is None:
                raise ValueError("account should not be None.")
            self.account.pack(packer)
            return
        if self.type == LedgerEntryType.TRUSTLINE:
            if self.trust_line is None:
                raise ValueError("trust_line should not be None.")
            self.trust_line.pack(packer)
            return
        if self.type == LedgerEntryType.OFFER:
            if self.offer is None:
                raise ValueError("offer should not be None.")
            self.offer.pack(packer)
            return
        if self.type == LedgerEntryType.DATA:
            if self.data is None:
                raise ValueError("data should not be None.")
            self.data.pack(packer)
            return
        if self.type == LedgerEntryType.CLAIMABLE_BALANCE:
            if self.claimable_balance is None:
                raise ValueError("claimable_balance should not be None.")
            self.claimable_balance.pack(packer)
            return
        if self.type == LedgerEntryType.LIQUIDITY_POOL:
            if self.liquidity_pool is None:
                raise ValueError("liquidity_pool should not be None.")
            self.liquidity_pool.pack(packer)
            return
        if self.type == LedgerEntryType.CONTRACT_DATA:
            if self.contract_data is None:
                raise ValueError("contract_data should not be None.")
            self.contract_data.pack(packer)
            return
        if self.type == LedgerEntryType.CONTRACT_CODE:
            if self.contract_code is None:
                raise ValueError("contract_code should not be None.")
            self.contract_code.pack(packer)
            return
        if self.type == LedgerEntryType.CONFIG_SETTING:
            if self.config_setting is None:
                raise ValueError("config_setting should not be None.")
            self.config_setting.pack(packer)
            return
        if self.type == LedgerEntryType.TTL:
            if self.ttl is None:
                raise ValueError("ttl should not be None.")
            self.ttl.pack(packer)
            return

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> LedgerEntryData:
        type = LedgerEntryType.unpack(unpacker)
        if type == LedgerEntryType.ACCOUNT:
            account = AccountEntry.unpack(unpacker)
            return cls(type=type, account=account)
        if type == LedgerEntryType.TRUSTLINE:
            trust_line = TrustLineEntry.unpack(unpacker)
            return cls(type=type, trust_line=trust_line)
        if type == LedgerEntryType.OFFER:
            offer = OfferEntry.unpack(unpacker)
            return cls(type=type, offer=offer)
        if type == LedgerEntryType.DATA:
            data = DataEntry.unpack(unpacker)
            return cls(type=type, data=data)
        if type == LedgerEntryType.CLAIMABLE_BALANCE:
            claimable_balance = ClaimableBalanceEntry.unpack(unpacker)
            return cls(type=type, claimable_balance=claimable_balance)
        if type == LedgerEntryType.LIQUIDITY_POOL:
            liquidity_pool = LiquidityPoolEntry.unpack(unpacker)
            return cls(type=type, liquidity_pool=liquidity_pool)
        if type == LedgerEntryType.CONTRACT_DATA:
            contract_data = ContractDataEntry.unpack(unpacker)
            return cls(type=type, contract_data=contract_data)
        if type == LedgerEntryType.CONTRACT_CODE:
            contract_code = ContractCodeEntry.unpack(unpacker)
            return cls(type=type, contract_code=contract_code)
        if type == LedgerEntryType.CONFIG_SETTING:
            config_setting = ConfigSettingEntry.unpack(unpacker)
            return cls(type=type, config_setting=config_setting)
        if type == LedgerEntryType.TTL:
            ttl = TTLEntry.unpack(unpacker)
            return cls(type=type, ttl=ttl)
        return cls(type=type)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> LedgerEntryData:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> LedgerEntryData:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.type,
                self.account,
                self.trust_line,
                self.offer,
                self.data,
                self.claimable_balance,
                self.liquidity_pool,
                self.contract_data,
                self.contract_code,
                self.config_setting,
                self.ttl,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.type == other.type
            and self.account == other.account
            and self.trust_line == other.trust_line
            and self.offer == other.offer
            and self.data == other.data
            and self.claimable_balance == other.claimable_balance
            and self.liquidity_pool == other.liquidity_pool
            and self.contract_data == other.contract_data
            and self.contract_code == other.contract_code
            and self.config_setting == other.config_setting
            and self.ttl == other.ttl
        )

    def __repr__(self):
        out = []
        out.append(f"type={self.type}")
        out.append(f"account={self.account}") if self.account is not None else None
        (
            out.append(f"trust_line={self.trust_line}")
            if self.trust_line is not None
            else None
        )
        out.append(f"offer={self.offer}") if self.offer is not None else None
        out.append(f"data={self.data}") if self.data is not None else None
        (
            out.append(f"claimable_balance={self.claimable_balance}")
            if self.claimable_balance is not None
            else None
        )
        (
            out.append(f"liquidity_pool={self.liquidity_pool}")
            if self.liquidity_pool is not None
            else None
        )
        (
            out.append(f"contract_data={self.contract_data}")
            if self.contract_data is not None
            else None
        )
        (
            out.append(f"contract_code={self.contract_code}")
            if self.contract_code is not None
            else None
        )
        (
            out.append(f"config_setting={self.config_setting}")
            if self.config_setting is not None
            else None
        )
        out.append(f"ttl={self.ttl}") if self.ttl is not None else None
        return f"<LedgerEntryData [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .int64 import Int64
from .uint32 import Uint32

__all__ = ["ConfigSettingContractEventsV0"]


class ConfigSettingContractEventsV0:
    """
    XDR Source Code::

        struct ConfigSettingContractEventsV0
        {
            // Maximum size of events that a contract call can emit.
            uint32 txMaxContractEventsSizeBytes;
            // Fee for generating 1KB of contract events.
            int64 feeContractEvents1KB;
        };
    """

    def __init__(
        self,
        tx_max_contract_events_size_bytes: Uint32,
        fee_contract_events1_kb: Int64,
    ) -> None:
        self.tx_max_contract_events_size_bytes = tx_max_contract_events_size_bytes
        self.fee_contract_events1_kb = fee_contract_events1_kb

    def pack(self, packer: Packer) -> None:
        self.tx_max_contract_events_size_bytes.pack(packer)
        self.fee_contract_events1_kb.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> ConfigSettingContractEventsV0:
        tx_max_contract_events_size_bytes = Uint32.unpack(unpacker)
        fee_contract_events1_kb = Int64.unpack(unpacker)
        return cls(
            tx_max_contract_events_size_bytes=tx_max_contract_events_size_bytes,
            fee_contract_events1_kb=fee_contract_events1_kb,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> ConfigSettingContractEventsV0:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> ConfigSettingContractEventsV0:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.tx_max_contract_events_size_bytes,
                self.fee_contract_events1_kb,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.tx_max_contract_events_size_bytes
            == other.tx_max_contract_events_size_bytes
            and self.fee_contract_events1_kb == other.fee_contract_events1_kb
        )

    def __repr__(self):
        out = [
            f"tx_max_contract_events_size_bytes={self.tx_max_contract_events_size_bytes}",
            f"fee_contract_events1_kb={self.fee_contract_events1_kb}",
        ]
        return f"<ConfigSettingContractEventsV0 [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from enum import IntEnum

from xdrlib3 import Packer, Unpacker

__all__ = ["CryptoKeyType"]


class CryptoKeyType(IntEnum):
    """
    XDR Source Code::

        enum CryptoKeyType
        {
            KEY_TYPE_ED25519 = 0,
            KEY_TYPE_PRE_AUTH_TX = 1,
            KEY_TYPE_HASH_X = 2,
            KEY_TYPE_ED25519_SIGNED_PAYLOAD = 3,
            // MUXED enum values for supported type are derived from the enum values
            // above by ORing them with 0x100
            KEY_TYPE_MUXED_ED25519 = 0x100
        };
    """

    KEY_TYPE_ED25519 = 0
    KEY_TYPE_PRE_AUTH_TX = 1
    KEY_TYPE_HASH_X = 2
    KEY_TYPE_ED25519_SIGNED_PAYLOAD = 3
    KEY_TYPE_MUXED_ED25519 = 256

    def pack(self, packer: Packer) -> None:
        packer.pack_int(self.value)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> CryptoKeyType:
        value = unpacker.unpack_int()
        return cls(value)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> CryptoKeyType:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> CryptoKeyType:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from typing import List

from xdrlib3 import Packer, Unpacker

from .contract_event import ContractEvent
from .sc_val import SCVal

__all__ = ["InvokeHostFunctionSuccessPreImage"]


class InvokeHostFunctionSuccessPreImage:
    """
    XDR Source Code::

        struct InvokeHostFunctionSuccessPreImage
        {
            SCVal returnValue;
            ContractEvent events<>;
        };
    """

    def __init__(
        self,
        return_value: SCVal,
        events: List[ContractEvent],
    ) -> None:
        _expect_max_length = 4294967295
        if events and len(events) > _expect_max_length:
            raise ValueError(
                f"The maximum length of `events` should be {_expect_max_length}, but got {len(events)}."
            )
        self.return_value = return_value
        self.events = events

    def pack(self, packer: Packer) -> None:
        self.return_value.pack(packer)
        packer.pack_uint(len(self.events))
        for events_item in self.events:
            events_item.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> InvokeHostFunctionSuccessPreImage:
        return_value = SCVal.unpack(unpacker)
        length = unpacker.unpack_uint()
        events = []
        for _ in range(length):
            events.append(ContractEvent.unpack(unpacker))
        return cls(
            return_value=return_value,
            events=events,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> InvokeHostFunctionSuccessPreImage:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> InvokeHostFunctionSuccessPreImage:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.return_value,
                self.events,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.return_value == other.return_value and self.events == other.events

    def __repr__(self):
        out = [
            f"return_value={self.return_value}",
            f"events={self.events}",
        ]
        return f"<InvokeHostFunctionSuccessPreImage [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from typing import List, Optional

from xdrlib3 import Packer, Unpacker

from .extension_point import ExtensionPoint
from .ledger_entry_changes import LedgerEntryChanges
from .operation_meta import OperationMeta
from .soroban_transaction_meta import SorobanTransactionMeta

__all__ = ["TransactionMetaV3"]


class TransactionMetaV3:
    """
    XDR Source Code::

        struct TransactionMetaV3
        {
            ExtensionPoint ext;

            LedgerEntryChanges txChangesBefore;  // tx level changes before operations
                                                 // are applied if any
            OperationMeta operations<>;          // meta for each operation
            LedgerEntryChanges txChangesAfter;   // tx level changes after operations are
                                                 // applied if any
            SorobanTransactionMeta* sorobanMeta; // Soroban-specific meta (only for
                                                 // Soroban transactions).
        };
    """

    def __init__(
        self,
        ext: ExtensionPoint,
        tx_changes_before: LedgerEntryChanges,
        operations: List[OperationMeta],
        tx_changes_after: LedgerEntryChanges,
        soroban_meta: Optional[SorobanTransactionMeta],
    ) -> None:
        _expect_max_length = 4294967295
        if operations and len(operations) > _expect_max_length:
            raise ValueError(
                f"The maximum length of `operations` should be {_expect_max_length}, but got {len(operations)}."
            )
        self.ext = ext
        self.tx_changes_before = tx_changes_before
        self.operations = operations
        self.tx_changes_after = tx_changes_after
        self.soroban_meta = soroban_meta

    def pack(self, packer: Packer) -> None:
        self.ext.pack(packer)
        self.tx_changes_before.pack(packer)
        packer.pack_uint(len(self.operations))
        for operations_item in self.operations:
            operations_item.pack(packer)
        self.tx_changes_after.pack(packer)
        if self.soroban_meta is None:
            packer.pack_uint(0)
        else:
            packer.pack_uint(1)
            self.soroban_meta.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> TransactionMetaV3:
        ext = ExtensionPoint.unpack(unpacker)
        tx_changes_before = LedgerEntryChanges.unpack(unpacker)
        length = unpacker.unpack_uint()
        operations = []
        for _ in range(length):
            operations.append(OperationMeta.unpack(unpacker))
        tx_changes_after = LedgerEntryChanges.unpack(unpacker)
        soroban_meta = (
            SorobanTransactionMeta.unpack(unpacker) if unpacker.unpack_uint() else None
        )
        return cls(
            ext=ext,
            tx_changes_before=tx_changes_before,
            operations=operations,
            tx_changes_after=tx_changes_after,
            soroban_meta=soroban_meta,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> TransactionMetaV3:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> TransactionMetaV3:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.ext,
                self.tx_changes_before,
                self.operations,
                self.tx_changes_after,
                self.soroban_meta,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.ext == other.ext
            and self.tx_changes_before == other.tx_changes_before
            and self.operations == other.operations
            and self.tx_changes_after == other.tx_changes_after
            and self.soroban_meta == other.soroban_meta
        )

    def __repr__(self):
        out = [
            f"ext={self.ext}",
            f"tx_changes_before={self.tx_changes_before}",
            f"operations={self.operations}",
            f"tx_changes_after={self.tx_changes_after}",
            f"soroban_meta={self.soroban_meta}",
        ]
        return f"<TransactionMetaV3 [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .base import String
from .constants import *
from .sc_spec_type_def import SCSpecTypeDef

__all__ = ["SCSpecUDTStructFieldV0"]


class SCSpecUDTStructFieldV0:
    """
    XDR Source Code::

        struct SCSpecUDTStructFieldV0
        {
            string doc<SC_SPEC_DOC_LIMIT>;
            string name<30>;
            SCSpecTypeDef type;
        };
    """

    def __init__(
        self,
        doc: bytes,
        name: bytes,
        type: SCSpecTypeDef,
    ) -> None:
        self.doc = doc
        self.name = name
        self.type = type

    def pack(self, packer: Packer) -> None:
        String(self.doc, SC_SPEC_DOC_LIMIT).pack(packer)
        String(self.name, 30).pack(packer)
        self.type.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> SCSpecUDTStructFieldV0:
        doc = String.unpack(unpacker)
        name = String.unpack(unpacker)
        type = SCSpecTypeDef.unpack(unpacker)
        return cls(
            doc=doc,
            name=name,
            type=type,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> SCSpecUDTStructFieldV0:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> SCSpecUDTStructFieldV0:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.doc,
                self.name,
                self.type,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.doc == other.doc
            and self.name == other.name
            and self.type == other.type
        )

    def __repr__(self):
        out = [
            f"doc={self.doc}",
            f"name={self.name}",
            f"type={self.type}",
        ]
        return f"<SCSpecUDTStructFieldV0 [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from enum import IntEnum

from xdrlib3 import Packer, Unpacker

__all__ = ["BucketEntryType"]


class BucketEntryType(IntEnum):
    """
    XDR Source Code::

        enum BucketEntryType
        {
            METAENTRY =
                -1, // At-and-after protocol 11: bucket metadata, should come first.
            LIVEENTRY = 0, // Before protocol 11: created-or-updated;
                           // At-and-after protocol 11: only updated.
            DEADENTRY = 1,
            INITENTRY = 2 // At-and-after protocol 11: only created.
        };
    """

    METAENTRY = -1
    LIVEENTRY = 0
    DEADENTRY = 1
    INITENTRY = 2

    def pack(self, packer: Packer) -> None:
        packer.pack_int(self.value)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> BucketEntryType:
        value = unpacker.unpack_int()
        return cls(value)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> BucketEntryType:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> BucketEntryType:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from enum import IntEnum

from xdrlib3 import Packer, Unpacker

__all__ = ["EndSponsoringFutureReservesResultCode"]


class EndSponsoringFutureReservesResultCode(IntEnum):
    """
    XDR Source Code::

        enum EndSponsoringFutureReservesResultCode
        {
            // codes considered as "success" for the operation
            END_SPONSORING_FUTURE_RESERVES_SUCCESS = 0,

            // codes considered as "failure" for the operation
            END_SPONSORING_FUTURE_RESERVES_NOT_SPONSORED = -1
        };
    """

    END_SPONSORING_FUTURE_RESERVES_SUCCESS = 0
    END_SPONSORING_FUTURE_RESERVES_NOT_SPONSORED = -1

    def pack(self, packer: Packer) -> None:
        packer.pack_int(self.value)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> EndSponsoringFutureReservesResultCode:
        value = unpacker.unpack_int()
        return cls(value)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> EndSponsoringFutureReservesResultCode:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> EndSponsoringFutureReservesResultCode:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .set_options_result_code import SetOptionsResultCode

__all__ = ["SetOptionsResult"]


class SetOptionsResult:
    """
    XDR Source Code::

        union SetOptionsResult switch (SetOptionsResultCode code)
        {
        case SET_OPTIONS_SUCCESS:
            void;
        case SET_OPTIONS_LOW_RESERVE:
        case SET_OPTIONS_TOO_MANY_SIGNERS:
        case SET_OPTIONS_BAD_FLAGS:
        case SET_OPTIONS_INVALID_INFLATION:
        case SET_OPTIONS_CANT_CHANGE:
        case SET_OPTIONS_UNKNOWN_FLAG:
        case SET_OPTIONS_THRESHOLD_OUT_OF_RANGE:
        case SET_OPTIONS_BAD_SIGNER:
        case SET_OPTIONS_INVALID_HOME_DOMAIN:
        case SET_OPTIONS_AUTH_REVOCABLE_REQUIRED:
            void;
        };
    """

    def __init__(
        self,
        code: SetOptionsResultCode,
    ) -> None:
        self.code = code

    def pack(self, packer: Packer) -> None:
        self.code.pack(packer)
        if self.code == SetOptionsResultCode.SET_OPTIONS_SUCCESS:
            return
        if self.code == SetOptionsResultCode.SET_OPTIONS_LOW_RESERVE:
            return
        if self.code == SetOptionsResultCode.SET_OPTIONS_TOO_MANY_SIGNERS:
            return
        if self.code == SetOptionsResultCode.SET_OPTIONS_BAD_FLAGS:
            return
        if self.code == SetOptionsResultCode.SET_OPTIONS_INVALID_INFLATION:
            return
        if self.code == SetOptionsResultCode.SET_OPTIONS_CANT_CHANGE:
            return
        if self.code == SetOptionsResultCode.SET_OPTIONS_UNKNOWN_FLAG:
            return
        if self.code == SetOptionsResultCode.SET_OPTIONS_THRESHOLD_OUT_OF_RANGE:
            return
        if self.code == SetOptionsResultCode.SET_OPTIONS_BAD_SIGNER:
            return
        if self.code == SetOptionsResultCode.SET_OPTIONS_INVALID_HOME_DOMAIN:
            return
        if self.code == SetOptionsResultCode.SET_OPTIONS_AUTH_REVOCABLE_REQUIRED:
            return

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> SetOptionsResult:
        code = SetOptionsResultCode.unpack(unpacker)
        if code == SetOptionsResultCode.SET_OPTIONS_SUCCESS:
            return cls(code=code)
        if code == SetOptionsResultCode.SET_OPTIONS_LOW_RESERVE:
            return cls(code=code)
        if code == SetOptionsResultCode.SET_OPTIONS_TOO_MANY_SIGNERS:
            return cls(code=code)
        if code == SetOptionsResultCode.SET_OPTIONS_BAD_FLAGS:
            return cls(code=code)
        if code == SetOptionsResultCode.SET_OPTIONS_INVALID_INFLATION:
            return cls(code=code)
        if code == SetOptionsResultCode.SET_OPTIONS_CANT_CHANGE:
            return cls(code=code)
        if code == SetOptionsResultCode.SET_OPTIONS_UNKNOWN_FLAG:
            return cls(code=code)
        if code == SetOptionsResultCode.SET_OPTIONS_THRESHOLD_OUT_OF_RANGE:
            return cls(code=code)
        if code == SetOptionsResultCode.SET_OPTIONS_BAD_SIGNER:
            return cls(code=code)
        if code == SetOptionsResultCode.SET_OPTIONS_INVALID_HOME_DOMAIN:
            return cls(code=code)
        if code == SetOptionsResultCode.SET_OPTIONS_AUTH_REVOCABLE_REQUIRED:
            return cls(code=code)
        return cls(code=code)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> SetOptionsResult:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> SetOptionsResult:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash((self.code,))

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.code == other.code

    def __repr__(self):
        out = []
        out.append(f"code={self.code}")
        return f"<SetOptionsResult [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from enum import IntEnum

from xdrlib3 import Packer, Unpacker

__all__ = ["EnvelopeType"]


class EnvelopeType(IntEnum):
    """
    XDR Source Code::

        enum EnvelopeType
        {
            ENVELOPE_TYPE_TX_V0 = 0,
            ENVELOPE_TYPE_SCP = 1,
            ENVELOPE_TYPE_TX = 2,
            ENVELOPE_TYPE_AUTH = 3,
            ENVELOPE_TYPE_SCPVALUE = 4,
            ENVELOPE_TYPE_TX_FEE_BUMP = 5,
            ENVELOPE_TYPE_OP_ID = 6,
            ENVELOPE_TYPE_POOL_REVOKE_OP_ID = 7,
            ENVELOPE_TYPE_CONTRACT_ID = 8,
            ENVELOPE_TYPE_SOROBAN_AUTHORIZATION = 9
        };
    """

    ENVELOPE_TYPE_TX_V0 = 0
    ENVELOPE_TYPE_SCP = 1
    ENVELOPE_TYPE_TX = 2
    ENVELOPE_TYPE_AUTH = 3
    ENVELOPE_TYPE_SCPVALUE = 4
    ENVELOPE_TYPE_TX_FEE_BUMP = 5
    ENVELOPE_TYPE_OP_ID = 6
    ENVELOPE_TYPE_POOL_REVOKE_OP_ID = 7
    ENVELOPE_TYPE_CONTRACT_ID = 8
    ENVELOPE_TYPE_SOROBAN_AUTHORIZATION = 9

    def pack(self, packer: Packer) -> None:
        packer.pack_int(self.value)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> EnvelopeType:
        value = unpacker.unpack_int()
        return cls(value)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> EnvelopeType:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> EnvelopeType:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)
